/*
 * $Id: simdef_copy.h,v 1.1 2004/07/02 09:06:08 stefans Exp $
 *
 * Primitives that may be used in test programs running on a simulated
 * CPU or virtual CPU.
 *
 */

#ifndef simdef_h
#define simdef_h

// --------------------------------------------------------------------- TARGET
#undef TARGET

#ifdef SIMDEF_TARGET_BANANA
#define TARGET banana
#define HAS_DMA
#endif

#ifdef SIMDEF_TARGET_SYSTEM
#define TARGET system
#define HAS_DMA
#endif

#ifdef SIMDEF_TARGET_MODEL
#define TARGET model
#endif

#ifdef USE_LIBC
#define TARGET system
#define SIMDEF_TARGET_SYSTEM
#define HAS_DMA
#endif

#ifndef TARGET
#error "SIMDEF_TARGET_BANANA, SIMDEF_TARGET_SYSTEM or SIMDEF_TARGET_MODEL must be defined"
#endif


// ----------------------------------------------------------------------- 
/* workaround for interrupt handlers in sos */
#ifdef SIMDEF_TARGET_SYSTEM
#define INTR_NEEDS_THIS_FIX 1
#endif
// ----------------------------------------------------------------------- libc

#ifdef USE_LIBC
#include <stdlib.h>
#include <syscall.h>
#include <vcpu.h>
#include <reg_map.h>
#include <dma.h>
#else
typedef unsigned size_t;
#endif // USE_LIBC

#ifndef SIMDEF_TARGET_MODEL

#ifndef USE_LIBC
/* libc-ish stuff */
extern struct FILE *stdin, *stdout, *stderr;
int  printf( const char* format, ... );
int  fprintf( struct FILE* stream, const char* format, ...);
#endif // ifndef USE_LIBC
void print_str( const char* str );
void print_hex( unsigned num );
void print_int( unsigned num );
unsigned virt2phys(unsigned addr);
unsigned phys2virt(unsigned addr);

#ifdef SIMDEF_TARGET_BANANA
#define USLEEP( t ) CPU_SLEEP( 1000*(t) )
#endif
#ifdef SIMDEF_TARGET_SYSTEM
#define CPU_INIT_INT()
#define USLEEP( t ) usleep( t )
#endif

#endif // ifndef SIMDEF_TARGET_MODEL

#ifndef SIMDEF_TARGET_MODEL

// -------------------------------------------------------------- multi_tasking
void yield();
void preempt_off();
void preempt_on();
int  preempt_is_off();

#ifndef USE_LIBC
#define PREEMPT_OFF_SAVE() do { int ____poff = preempt_is_off(); \
                                if( !____poff ) preempt_off()
#define PREEMPT_RESTORE()       if( !____poff ) preempt_on(); } while( 0 )
#endif

// ------------------------------------------------------------------------ DMA
// defines descriptors and registers and a few DMA macros
#ifdef HAS_DMA
#include "dma.h"
/* defines these macros:
   DMA_RESET( inst )
   DMA_START_GROUP( inst, group_descr )
   DMA_START_CONTEXT( inst, context_descr )
   DMA_STOP( inst )
   DMA_RESTART( inst )
*/

void *dma_start_simple( unsigned inst, void *data, unsigned size );

void *dma_start_simple_out( unsigned inst, void *data, unsigned size );
int   dma_start_simple_in(  unsigned inst, void *data, unsigned size );

#endif // HAS_DMA

// -------------------------------------------------------------- shared_memory

void *malloc( size_t size );
void *calloc( size_t nmemb, size_t size );
void free( void *ptr );
void *realloc( void *ptr, size_t size );

#undef  RAND_MAX
#define RAND_MAX 0x7fffffff
int rand( void );
void srand( unsigned int seed );

int rand_range( int min, int max );


// ----------------------------------------------------------------------- VCPU
/* VCPU stuff, TARGET specific include file: */
#include "vcpu.h"

/* Defines these macros:
   #define CPU_SMIF_RD( addr, smif_d, size)
   #define CPU_SMIF_WR( addr, smif_d, size)
   #define CPU_SMIF_ABORT_RD( addr, smif_d, size)
   #define CPU_SMIF_ABORT_WR( addr, smif_d, size)
   #define CPU_SET_VMEM_BASE( addr, vmem_nbr)
   #define CPU_SLEEP( ns )
   #define CPU_STOP_SIM()
   #define CPU_INIT_INT()
   #define CPU_SET_INT_VEC( vecnr, addr )
   #define CPU_ENABLE_INT_VEC( vecnr)
   #define CPU_DISABLE_INT_VEC( vecnr)
   #define CPU_IS_ENABLED_INT_VEC( vecnr )
   #define CPU_GET_SIM_TIME()
   #define CPU_KICK_DOG()
   #define CPU_WATCHDOG_TIMEOUT( us )
*/

// ---------------------------------------------------------------- Help_macros
/* Help_macros */

/* Yes, the two levels of STRINGIFY is actually needed ... */
#define STRINGIFYFY(i) #i
#define STRINGIFY(i) STRINGIFYFY(i)

#define ERRPOS " @" __FILE__ ":" STRINGIFY(__LINE__)
#define PERROR( arg ) print_str( "\nERROR: " arg )
#define UNSUPPORTED( name ) do { PERROR( #name " not supported for " \
                                         TARGET " test-benches\n" ); \
                                 CPU_STOP_SIM(); \
                               } while( 0 )


// ----------------------------------------------------------------------- TEST
/* test define/control stuff */

extern char *test_module, *test_name;

/* Start of test: */
#define TEST( name ) do { \
                    PREEMPT_OFF_SAVE(); \
                    test_name = name; \
                    print_str( "\nstart test: " ); \
                    print_str( test_module ); \
                    print_str( "." name ERRPOS "\n"); \
		    PREEMPT_RESTORE(); \
                    } while ( 0 )

/* Test PASSED */
#define PASS() do { \
                    PREEMPT_OFF_SAVE(); \
		    print_str( "\ntest passed: " ); \
                    print_str( test_module ); \
                    print_str( "." ); \
                    print_str( test_name ); \
                    print_str( ERRPOS ); \
                    print_str( "\n" ); \
                    test_name = "<none>"; \
		    PREEMPT_RESTORE(); \
                  } while( 0 )

/* Test FAILED */
#define FAIL() do { \
                    PREEMPT_OFF_SAVE(); \
                    print_str( "\ntest failed: " ); \
                    print_str( test_module ); \
                    print_str( "." ); \
                    print_str( test_name ); \
                    print_str( ERRPOS ); \
                    print_str( "\n" ); \
                    CPU_STOP_SIM(); \
		    PREEMPT_RESTORE(); \
                  } while( 0 )

/* Fail if cond is false */
#define ASSERT( cond ) do { if( !(cond) ) { \
                              PERROR( "assertion failed: " #cond ); \
                              FAIL(); \
                            } \
                          } while( 0 )

/* init_<module> stuff */
#define TEST_MODULE( name ) do { test_module = name; \
                                 print_str( "\ntest module: " name "\n" ); \
                               } while( 0 )


#define TEST_MVP( path ) print_str( "\nmvp: " path "\n" );


#define TEST_REG_INST_DECL( name, value ) enum { name = value }
#define TEST_IRQ_INST_DECL( name, value ) enum { name = value }


// --------------------------------------------------------------------- memory

void *alloc_random_data( unsigned min_size,
                         unsigned max_size,
                         unsigned align,         // align to 2**align
                         unsigned *seed,
                         unsigned *csize );

void *calloc_random_data( unsigned min_size,
                          unsigned max_size,
                          unsigned align,
                          unsigned *seed,
                          unsigned *csize );


#endif // ifndef SIMDEF_TARGET_MODEL

void print_mem_hex( void *ptr, size_t size );


typedef unsigned int shmem_handle;

shmem_handle   alloc_shared( size_t size );
void          *ptr_shared( shmem_handle handle );
void           free_shared( shmem_handle handle );
shmem_handle   file2shared( char *filename, size_t *size );
int            cmp_shared( shmem_handle handle1, shmem_handle handle2, 
                           size_t size );
int            cmp_offset_shared( shmem_handle handle1, unsigned offset1,
                                  shmem_handle handle2, unsigned offset2,
                                  size_t size );
void           fill_random_shared( shmem_handle handle1, unsigned offset,
                                   size_t size, unsigned *seed );

#ifndef SIMDEF_TARGET_MODEL
void           print_descr_context(dma_descr_context *descr);
void           print_descr_data(dma_descr_data *descr);
void           print_descr_group(dma_descr_group *descr);
#endif


#endif // ifndef simdef_h
// ------------------------------------------------------------------------ eof
