/* Generate a ROMFS file system
 *
 * Copyright (C) 1997  Janos Farkas <chexum@shadow.banki.hu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * Changes:
 *	2 Jan 1997				Initial release
 *      6 Aug 1997				Second release
 *
 * Some sparse words about how to use the program
 *
 * `genromfs' is the `mkfs' equivalent of the other filesystems, but
 * you must tell it from which directory you want to build the
 * filesystem.  I.e. all files (and directories) in that directory
 * will be part of the newly created filesystem.  Imagine it like
 * building a cd image, or creating an archive (tar, zip) file.
 *
 * Basic usage:
 *
 * # genromfs -d rescue/ -f /dev/fd0
 *
 * All files in the rescue directory will be written to /dev/fd0 as a
 * new romfs filesystem image.  You can mount it (if you have the
 * romfs module loaded, or compiled into the kernel) via:
 *
 * # mount -t romfs /dev/fd0 /mnt
 *
 * You can also set the volume name of the filesystem (which is not
 * currently used by the kernel) with the -V option.  If you don't
 * specify one, genromfs will create a volume name of the form: 'rom
 * xxxxxxxx', where the x's represent the current time in a cryptic
 * form.
 *
 * All in all, it's as simple as:
 *
 * # genromfs -d rescue -f testimg.rom -V "Install disk"
 *
 * Warning!  Quite spaghetti code, it was born in a few hours.
 * Sorry about that.  Feel free to contact me if you have problems.
 *
 * $Id: genromfs.c,v 1.10 2000/12/08 20:28:18 tobiasa Exp $
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/time.h>
#include <netinet/in.h>

#ifdef __linux__
#include <sys/sysmacros.h>
#endif /* __linux__ */

#if 0
/* we run genromfs in a little-endian host (x86) and Etrax is little-endian as
   well, so we dont want to store the filesystem in big-endian format (network)
*/

#define ntohl(x)        (x)
#define ntohs(x)        (x)
#define htonl(x)        (x)
#define htons(x)        (x)
#endif

typedef mode_t umode_t;

struct romfh {
	int nextfh;
	int spec;
	int size;
	int checksum;
};

#define ROMFS_MAXFN 128
#define ROMFH_HRD 0
#define ROMFH_DIR 1
#define ROMFH_REG 2
#define ROMFH_LNK 3
#define ROMFH_BLK 4
#define ROMFH_CHR 5
#define ROMFH_SCK 6
#define ROMFH_FIF 7
#define ROMFH_EXEC 8

struct filenode;

struct filehdr {
	struct filenode *head;
	struct filenode *tail;
	struct filenode *tailpred;
};

struct filenode {
	struct filenode *next;
	struct filenode *prev;
	struct filehdr dirlist;
	struct filenode *orig_link;
	char *name;
	char *realname;
	dev_t ondev;
	dev_t devnode;
	ino_t onino;
	umode_t modes;
	unsigned int offset;
	unsigned int size;
};

/* Meta data functions */

#define META_IGNORE         1
#define META_IGNORECONTENTS 2
#define META_DEVICE         3
#define META_INCLUDE        4

struct metadata {
	struct metadata *next;
	int type;
	char *name;
	char dev_type;
	unsigned char major;
	unsigned char minor;
};

#define BUFFERLENGTH 256

struct metadata *create_meta_ignore(char *line, struct metadata *meta, int ignoretype)
{
	struct metadata *new_meta;
	char *end = &line[strlen(line)];

	while (isspace(*line)) {
		line++;
	}

	for (end--; end > line && isspace(*end); end--) {
		*end = '\0';
	}

	if (end > line && *line == '"' && *end == '"') {
		line++;
		*end = '\0';
	}

	/* Can't ignore the . and .. directories */
	if (strlen(line) && strcmp(line, ".") && strcmp(line, "..")) {
		if ((new_meta = malloc(sizeof(struct metadata)))) {
			new_meta->next = meta;
			new_meta->type = ignoretype;
			new_meta->name = strdup(line);
			return new_meta;
		} else {
			fprintf(stderr,"out of memory\n");
			exit(1);
		}
	}

	return meta;
}

struct metadata *create_meta_include(char *line, struct metadata *meta)
{
	struct metadata *new_meta;
	char *end = &line[strlen(line)];

	while (isspace(*line)) {
		line++;
	}

	for (end--; end > line && isspace(*end); end--) {
		*end = '\0';
	}

	if (end > line && *line == '"' && *end == '"') {
		line++;
		*end = '\0';
	}

	if ((new_meta = malloc(sizeof(struct metadata)))) {
		new_meta->next = meta;
		new_meta->type = META_INCLUDE;
		new_meta->name = strdup(line);
		return new_meta;
	} else {
		fprintf(stderr,"out of memory\n");
		exit(1);
	}

	return meta;
}

struct metadata *create_meta_device(char *line, struct metadata *meta)
{
	struct metadata *new_meta;
	char name[100];
	char dev_type;
	unsigned int major = 0;
	unsigned int minor = 0;
	int found;

	found = sscanf(line, "%s %c %u %u", name, &dev_type, &major, &minor);
	
	if ((found == 2 || found == 4) &&
	    (dev_type == 'b' || dev_type == 'c' || dev_type == 'p') &&
	    major < 256 && minor < 256) {
		if ((new_meta = malloc(sizeof(struct metadata)))) {
			new_meta->next = meta;
			new_meta->type = META_DEVICE;
			new_meta->name = strdup(name);
			new_meta->dev_type = dev_type;
			if (dev_type == 'b' || dev_type == 'c') {
				new_meta->major = (unsigned char)major;
				new_meta->minor = (unsigned char)minor;
			}
			return new_meta;
		} else {
			fprintf(stderr,"out of memory\n");
			exit(1);
		}
	} else
		fprintf(stderr, "Illegal device file specification: %s", line);

	return meta;
}

struct metadata *read_meta_data(char *dir_name)
{
	FILE *meta_file;
	char *meta_file_name;
	struct metadata *meta = 0;
	char line[BUFFERLENGTH];

	if (!(meta_file_name = malloc(strlen(dir_name) + 16))) {
		return 0;
	}

	strcpy(meta_file_name, dir_name);
	strcat(meta_file_name, "/romfs_meta.txt");

	if (!(meta_file = fopen(meta_file_name, "r"))) {
		return 0;
	}

	while (!feof(meta_file)) {
		if (fgets(line, BUFFERLENGTH-1, meta_file)) {
			if (!strncasecmp("Ignore:", line, 7)) {
				meta = create_meta_ignore(line+7, meta, META_IGNORE);
			} else if (!strncasecmp("Ignorecontents:", line, 15)) {
				meta = create_meta_ignore(line+15, meta,
							  META_IGNORECONTENTS);
			} else if (!strncasecmp("Include:", line, 8)) {
				meta = create_meta_include(line+8, meta);
			} else if (!strncasecmp("Device:", line, 7)) {
				meta = create_meta_device(line+7, meta);
			}
		}
	}
	
	fclose(meta_file);
	free(meta_file_name);

	return meta;
}

void free_meta_data(struct metadata *meta)
{
	struct metadata *next;

	for (; meta; meta = next) {
		next = meta->next;
		free(meta->name);
		free(meta);
	}
}

int meta_ignore(struct metadata *meta, char *name)
{
	for(; meta; meta = meta->next) {
		if (meta->type == META_IGNORE && !strcmp(name, meta->name)) {
			return 1;
		}
	}

	return 0;
}

int meta_ignorecontents(struct metadata *meta, char *name)
{
	for(; meta; meta = meta->next) {
		if (meta->type == META_IGNORECONTENTS && !strcmp(name, meta->name)) {
			return 1;
		}
	}

	return 0;
}

int meta_include(struct metadata *meta, char *name)
{
	int found = 0;

	/* Always include . and .. */
	if (!strcmp(name, ".") || !strcmp(name, ".."))
		return 1;  

	for(; meta; meta = meta->next) {
		if (meta->type == META_INCLUDE) {
			found = 1;
			if (!strcmp(name, meta->name)) {
				return 1;
			}
		}
	}

	return !found;
}

int meta_handle_device(struct metadata *meta, struct filenode *node)
{
	for(; meta; meta = meta->next) {
		if (meta->type == META_DEVICE &&
		    !strcmp(node->name, meta->name)) {
			if (meta->dev_type == 'b') {
				node->modes =
					(node->modes & ~S_IFMT) | S_IFBLK;
				node->devnode =
					makedev(meta->major, meta->minor);
			} else if (meta->dev_type == 'c') {
				node->modes =
					(node->modes & ~S_IFMT) | S_IFCHR;
				node->devnode =
					makedev(meta->major, meta->minor);
			} else {
				node->modes =
					(node->modes & ~S_IFMT) | S_IFIFO;
			}
			return 1;
		}
	}

	return 0;
}

void initlist(struct filehdr *fh)
{
	fh->head = (struct filenode *)&fh->tail;
	fh->tail = NULL;
	fh->tailpred = (struct filenode *)&fh->head;
}

int listisempty(struct filehdr *fh)
{
	return fh->head == (struct filenode *)&fh->tail;
}

void append(struct filehdr *fh, struct filenode *n)
{
	struct filenode *tail = (struct filenode *)&fh->tail;

	n->next = tail; n->prev = tail->prev;
	tail->prev = n; n->prev->next =n;
}

void shownode(int level, struct filenode *node, FILE *f)
{
	struct filenode *p;

	fprintf(f, "%-4d %-20s [0x%-8x, 0x%-8x] %07o, sz %5u, at 0x%-6x", level, node->name,
		(int)node->ondev, (int)node->onino,
		node->modes,
		node->size, node->offset);
	if (S_ISCHR(node->modes) || S_ISBLK(node->modes))
		fprintf(f, " [maj %d, min %d]", major(node->devnode), minor(node->devnode));
	if (node->orig_link)
		fprintf(f, " [link to 0x%-6x]", node->orig_link->offset);
	fprintf(f, "\n");
	p = node->dirlist.head;
	while (p->next) {
		shownode(level+1, p, f);
		p = p->next;
	}
}

/* Dumping functions */

static char bigbuf[4096];
static char fixbuf[512];
static int atoffs = 0;

int romfs_checksum(void *data, int size)
{
        int sum, *ptr;

        sum = 0; ptr = data;
        size>>=2;
        while (size>0) {
                sum += htonl(*ptr++);
                size--;
        }
        return sum;
}

void fixsum(struct romfh *ri, int size)
{
	ri->checksum = 0;
	ri->checksum = ntohl(-romfs_checksum(ri, size));
}

void dumpdata(void *addr, int len, FILE *f)
{
	int tocopy;
	struct romfh *ri;

	if (!len)
		return;
	if (atoffs >= 512) {
		fwrite(addr, len, 1, f);
		atoffs += len;
		return;
	}

	tocopy = len < 512-atoffs ? len : 512-atoffs;
	memcpy(fixbuf+atoffs, addr, tocopy);
	atoffs += tocopy;
	addr = (char*)addr+tocopy;
	len -= tocopy;

	if (atoffs == 512) {
		ri = (struct romfh *)&fixbuf;
		fixsum(ri, atoffs < ntohl(ri->size) ? atoffs : ntohl(ri->size));
		fwrite(fixbuf, atoffs, 1, f);
	}
	if (len) {
		fwrite(addr, len, 1, f);
		atoffs += len;
	}
}

void dumpzero(int len, FILE *f)
{
	memset(bigbuf, 0, len);
	dumpdata(bigbuf, len, f);
}

void dumpdataa(void *addr, int len, FILE *f)
{
	dumpdata(addr, len, f);
	if ((len & 15) != 0)
		dumpzero(16-(len&15), f);
}

void dumpstring(char *str, FILE *f)
{
	dumpdataa(str, strlen(str)+1, f);
}

void dumpri(struct romfh *ri, struct filenode *n, FILE *f)
{
	int len;

	len = strlen(n->name)+1;
	memcpy(bigbuf, ri, sizeof(*ri));
	memcpy(bigbuf+16, n->name, len);
	if (len&15) {
		memset(bigbuf+16+len, 0, 16-(len&15));
		len += 16-(len&15);
	}
	len+=16;
	ri=(struct romfh *)bigbuf;
	if (n->offset)
		fixsum(ri, len);
	dumpdata(bigbuf, len, f);
#if 0
	fprintf(stderr, "RI: [at %06x] %08lx, %08lx, %08lx, %08lx [%s]\n",
		n->offset,
		ntohl(ri->nextfh), ntohl(ri->spec),
		ntohl(ri->size), ntohl(ri->checksum),
		n->name);
#endif
}

void dumpnode(struct filenode *node, FILE *f)
{
	struct romfh ri;
	struct filenode *p;

	ri.nextfh = 0;
	ri.spec = 0;
	ri.size = htonl(node->size);
	ri.checksum = htonl(0x55555555);
	if (node->next && node->next->next)
		ri.nextfh = ntohl(node->next->offset);
	if ((node->modes & 0111) &&
	    (S_ISDIR(node->modes) || S_ISREG(node->modes)))
		ri.nextfh |= ntohl(ROMFH_EXEC);

	if (node->orig_link) {
		ri.nextfh |= ntohl(ROMFH_HRD);
		/* Don't allow hardlinks to convey attributes */
		ri.nextfh &= ~ntohl(ROMFH_EXEC);
		ri.spec = ntohl(node->orig_link->offset);
		dumpri(&ri, node, f);
	} else if (S_ISDIR(node->modes)) {
		ri.nextfh |= ntohl(ROMFH_DIR);
		if (listisempty(&node->dirlist)) {
			ri.spec = ntohl(node->offset);
		} else {
			ri.spec = ntohl(node->dirlist.head->offset);
		}
		dumpri(&ri, node, f);
	} else if (S_ISLNK(node->modes)) {
		ri.nextfh |= ntohl(ROMFH_LNK);
		dumpri(&ri, node, f);
		memset(bigbuf, 0, sizeof(bigbuf));
		readlink(node->realname, bigbuf, node->size);
		dumpdataa(bigbuf, node->size, f);
	} else if (S_ISREG(node->modes)) {
		int offset, len, fd, max, avail;
		ri.nextfh |= ntohl(ROMFH_REG);
		dumpri(&ri, node, f);
		offset = 0;
		max = node->size;
		/* XXX warn about size mismatch */
		fd = open(node->realname, O_RDONLY);
		if (fd) {
			while(offset < max) {
				avail = max-offset < sizeof(bigbuf) ? max-offset : sizeof(bigbuf);
				len = read(fd, bigbuf, avail);
				if (len <= 0)
					break;
				dumpdata(bigbuf, len, f);
				offset+=len;
			}
			close(fd);
		}
		max = (max+15)&~15;
		while (offset < max) {
			avail = max-offset < sizeof(bigbuf) ? max-offset : sizeof(bigbuf);
			memset(bigbuf, '\0', avail);
			dumpdata(bigbuf, avail, f);
			offset+=avail;
		}
	} else if (S_ISCHR(node->modes)) {
		ri.nextfh |= ntohl(ROMFH_CHR);
		ri.spec = ntohl(major(node->devnode)<<16|minor(node->devnode));
		dumpri(&ri, node, f);
	} else if (S_ISBLK(node->modes)) {
		ri.nextfh |= ntohl(ROMFH_BLK);
		ri.spec = ntohl(major(node->devnode)<<16|minor(node->devnode));
		dumpri(&ri, node, f);
	} else if (S_ISFIFO(node->modes)) {
		ri.nextfh |= ntohl(ROMFH_FIF);
		dumpri(&ri, node, f);
	} else if (S_ISSOCK(node->modes)) {
		ri.nextfh |= ntohl(ROMFH_SCK);
		dumpri(&ri, node, f);
	}

	p = node->dirlist.head;
	while (p->next) {
		dumpnode(p, f);
		p = p->next;
	}
}

void dumpall(struct filenode *node, int lastoff, FILE *f)
{
	struct romfh ri;
	struct filenode *p;

	ri.nextfh = htonl(0x2d726f6d);
	ri.spec = htonl(0x3166732d);
	ri.size = htonl(lastoff);
	ri.checksum = htonl(0x55555555);
	dumpri(&ri, node, f);
	p = node->dirlist.head;
	while (p->next) {
		dumpnode(p, f);
		p = p->next;
	}
	/* Align the whole bunch to ROMBSIZE boundary */
	if (lastoff&1023)
		dumpzero(1024-(lastoff&1023), f);
}

/* Node manipulating functions */

void freenode(struct filenode *n)
{
	/* Rare, not yet */
}

void setnode(struct filenode *n, dev_t dev, ino_t ino, umode_t um)
{
	n->ondev = dev;
	n->onino = ino;
	n->modes = um;
}

struct filenode *newnode(const char *base, const char *name, int curroffset)
{
	struct filenode *node;
	int len;
	char *str;

	node = malloc(sizeof (*node));
	if (!node) {
		fprintf(stderr,"out of memory\n");
		exit(1);
	}

	len = strlen(name);
	str = malloc(len+1);
	if (!str) {
		fprintf(stderr,"out of memory\n");
		exit(1);
	}
	strcpy(str, name);
	node->name = str;

	if (!curroffset) {
		len = 1;
		name = ".";
	}
	if (strlen(base))
		len++;
	str = malloc(strlen(base)+len+1);
	if (!str) {
		fprintf(stderr,"out of memory\n");
		exit(1);
	}
	if (strlen(base)) {
		sprintf(str, "%s/%s", base, name);
	} else {
		strcpy(str, name);
	}

	node->realname = str;
	node->next = node->prev = NULL;
	initlist(&node->dirlist);

	node->ondev = -1;
	node->onino = -1;
	node->modes = -1;
	node->size = 0;
	node->devnode = 0;
	node->orig_link = NULL;
	node->offset = curroffset;

	return node;
}

struct filenode *findnode(struct filenode *node, dev_t dev, ino_t ino)
{
	struct filenode *found, *p;

	/* scan the whole tree */
	if (node->ondev == dev && node->onino == ino)
		return node;
	p = node->dirlist.head;
	while (p->next) {
		found = findnode(p, dev, ino);
		if (found)
			return found;
		p = p->next;
	}
	return NULL;
}

#define ALIGNUP16(x) (((x)+15)&~15)

int spaceneeded(struct filenode *node)
{
	return 16 + ALIGNUP16(strlen(node->name)+1) + ALIGNUP16(node->size);
}

int processdir(int level, const char *base, const char *dirname,
               struct stat *sb, struct filenode *dir, struct filenode *root,
	       int curroffset, int ignorecontents)
{
	DIR *dirfd;
	struct dirent *dp;
	struct filenode *n, *link;
	struct metadata *meta;

	meta = read_meta_data(dir->realname);

	if (level <= 1) {
		/* Ok, to make sure . and .. are handled correctly
		 * we add them first.  Note also that we alloc them
		 * first to get to know the real name
		 */
		link = newnode(base, ".", curroffset);
		if (!lstat(link->realname, sb)) {
			setnode(link, sb->st_dev, sb->st_ino, sb->st_mode);
			append(&dir->dirlist, link);
			curroffset += spaceneeded(link);
			n = newnode(base, "..", curroffset);
			if (!lstat(n->realname, sb)) {
				setnode(n, sb->st_dev, sb->st_ino, sb->st_mode);
				append(&dir->dirlist, n);
				n->orig_link = link;
				curroffset += spaceneeded(n);
			}
		}
	}

	dirfd = opendir(dir->realname);
	while ((dp = readdir(dirfd))) {
		if (ignorecontents &&
                    strcmp(dp->d_name, ".") &&
                    strcmp(dp->d_name, ".."))
			continue; /* skip contents except . and .. ? */
		if (!strcmp(dp->d_name, "romfs_meta.txt"))
			continue; /* don't include the meta data file */
		if (!strcmp(dp->d_name, "CVS"))
			continue; /* don't include CVS directories */
		if (!strcmp(dp->d_name, ".cvsignore"))
			continue; /* don't include CVS ignore files */
		if (!strcmp(dp->d_name, ".keep_this"))
			continue; /* don't include .keep_this files, used */
				  /* to keep empty directories around */
		if (!meta_include(meta, dp->d_name))
			continue; /* include files according to the meta data */
		if (meta_ignore(meta, dp->d_name))
			continue; /* ignore files according to the meta data */
		/* don't process main . and .. twice */
		if (level <= 1 &&
		    (strcmp(dp->d_name, ".") == 0
		     || strcmp(dp->d_name, "..") == 0))
			continue;
		
		n = newnode(base, dp->d_name, curroffset);
		if (lstat(n->realname, sb)) {
			fprintf(stderr, "ignoring '%s' (lstat failed)\n", n->realname);
			freenode(n); continue;
		}
		setnode(n, sb->st_dev, sb->st_ino, sb->st_mode);
		/* Skip unreadable files/dirs */
		if (!S_ISLNK(n->modes) && access(n->realname, R_OK)) {
			fprintf(stderr, "ignoring '%s' (access failed)\n", n->realname);
			freenode(n); continue;
		}
		/* Look up old links */
		link = findnode(root, n->ondev, n->onino);
		append(&dir->dirlist, n);
		if (link) {
			n->orig_link = link;
			curroffset += spaceneeded(n);
			continue;
		}
		/* possibly we override the node definition */
		if (meta_handle_device(meta, n))
			curroffset += spaceneeded(n);
		else {
			if (S_ISREG(sb->st_mode)) {
				n->size = sb->st_size;
			}
			if (S_ISLNK(sb->st_mode)) {
				n->size = sb->st_size;
			}
			curroffset += spaceneeded(n);
			if (S_ISCHR(sb->st_mode) || S_ISBLK(sb->st_mode)) {
				n->devnode = sb->st_rdev;
			}
			if (S_ISDIR(sb->st_mode)) {
				curroffset = processdir(level+1, n->realname,
							dp->d_name, sb, n, root,
							curroffset,
							meta_ignorecontents(meta,
									    dp->d_name));
			}
		}
	}
	closedir(dirfd);
	free_meta_data(meta);
	return curroffset;
}

void showhelp(const char *argv0)
{
	printf("Usage: %s [OPTIONS] -f IMAGE\n",argv0);
	printf("Create a romfs filesystem image from a directory\n");
	printf("\n");
	printf("  -f IMAGE               Output the image into this file\n");
	printf("  -d DIRECTORY           Use this directory as source\n");
	printf("  -v                     (Too) verbose operation\n");
	printf("  -V VOLUME              Use the specified volume name\n");
	printf("  -h                     Show this help\n");
	printf("\n");
	printf("Report bugs to chexum@shadow.banki.hu\n");
}

int main(int argc, char *argv[])
{
	char c;
	char *dir = ".";
	char *outf = NULL;
	char *volname = NULL;
	int verbose=0;
	char buf[256];
	struct filenode *root;
	struct stat sb;
	int lastoff;
	FILE *f;

	while ((c = getopt(argc, argv, "V:vd:f:h")) != EOF) {
		switch(c) {
		case 'd':
			dir = optarg;
			break;
		case 'f':
			outf = optarg;
			break;
		case 'V':
			volname = optarg;
			break;
		case 'v':
			verbose = 1;
			break;
		case 'h':
			showhelp(argv[0]);
			exit(0);
		default:
			exit(1);
		}
	}

	if (!volname) {
		sprintf(buf, "rom %08lx", time(NULL));
		volname = buf;
	}
	if (!outf) {
		fprintf(stderr, "%s: you must specify the destination file\n", argv[0]);
		fprintf(stderr, "Try `%s -h' for more information\n", argv[0]);
		exit(1);
	}
	if (strcmp(outf, "-") == 0) {
		f = fdopen(1, "w");
	} else
		f = fopen(outf, "w");

	if (!f) {
		perror(outf);
		exit(1);
	}

	root = newnode(dir, volname, 0);
	lastoff = processdir (1, dir, dir, &sb, root, root, spaceneeded(root), 0);
	if (verbose)
		shownode(0, root, stderr);
	dumpall(root, lastoff, f);

	exit(0);
}
