#!/usr/bin/perl -w

#*****************************************************************************
#!
#! FILE NAME  : prodspec.pm
#!
#! DESCRIPTION: Reads and parses a prodspec file.
#!
#! FUNCTIONS  : read_prodspec
#! (EXPORTED)
#!
#! FUNCTIONS  :
#! (LOCAL)
#!
#!----------------------------------------------------------------------------
#! HISTORY
#!
#! $Log: prodspec.pm,v $
#! Revision 1.6  2002/11/18 09:58:40  pkj
#! Implemented parse_ptable().
#!
#! Revision 1.5  2002/10/31 09:44:55  pkj
#! Added support for file options.
#!
#! Revision 1.4  2002/10/30 11:37:11  pkj
#! * Handle backticked commands.
#! * Handle sections that depend on the build type.
#! * Handle the swinfo and image sections.
#! * Handle the -s argument.
#! * Do not use a special section for comments.
#! * Do not automatically set the revision to HEAD if one was not present.
#! * Add quotes around arguments that contain spaces.
#!
#! Revision 1.3  2002/01/29 10:42:36  pkj
#! * Support most sections.
#! * Support branches.
#!
#! Revision 1.2  2001/03/16 15:45:43  pkj
#! Ignore config and build sections.
#!
#! Revision 1.1  2000/12/06 16:36:28  pkj
#! The beginning of a general prodspec parser. Currently only parses
#! the apps, libs, packages, source and tools sections, and it does
#! not handle commands within backticks (`).
#!
#!----------------------------------------------------------------------------
#! Copyright (C) 2000, Axis Communications AB, LUND, SWEDEN
#!****************************************************************************
# $Id: prodspec.pm,v 1.6 2002/11/18 09:58:40 pkj Exp $

#****************** PACKAGE SECTION ******************************************

package prodspec;
require Exporter;
@ISA       = qw(Exporter);

# These functions and variables are exported by default
@EXPORT    = qw(read_prodspec);

# These functions and variables may be exported on request
@EXPORT_OK = qw();

#****************** INCLUDE FILES SECTION ************************************

use strict;
no strict 'refs';

use Text::ParseWords;

#****************** VARIABLE DECLARATION SECTION *****************************

use vars qw();

#****************** CONSTANT SECTION *****************************************

#****************** FUNCTION DEFINITION SECTION ******************************

#*****************************************************************************
##
## FUNCTION NAME: read_prodspec
##
## PARAMETERS   : prodspec        - The path and name of the prodspec file
##                ignore_revision - Ignore the revisions specified in the
##                                  prodspec file
##
## RETURNS      : Returns undef if the prodspec file could not be opened.
##                Otherwise an array of hash references with the commands
##                specified in the prodspec file.
##                The following information may be available for each command:
##
##                section       - current section | 'comment'
##                file          - path (and file name)
##                full_file     - path (and file name) including apps, libs
##                                or packages
##                revision      - specified revision | 'HEAD'
##                branch        - specified branch
##                target        - specified target | 'install' | ''
##                noreleasenote - 1 if -noreleasenote specified
##                noautotarget  - 1 if -noautotarget specified
##                options       - all unparsed options
##                comment       - trailing comment
##
## SIDE EFFECTS : None.
##
## DESCRIPTION  : Reads and parses a prodspec file. Returns an array with
##                references to hashes that specify the commands listed
##                in the prodspec file.
##
##****************************************************************************

sub read_prodspec
{
  my($prodspec, $ignore_revision) = @_;
  my $section = 'shell';
  my $sub_section = '';
  my @lines;

  $ignore_revision ||= 0;

  return undef unless (open(PRODSPEC, "<$prodspec"));

  while (<PRODSPEC>)
  {
    my(@line);

    if (/^(\s*\#.*)/ || /^\s*$/)
    {
      # Comments and empty lines
      my %cmd;
      $cmd{'section'} = $section;
      $cmd{'sub_section'} = $sub_section;
      $cmd{'comment'} = $1 || '';
      push @lines, \%cmd;
    }
    elsif (/^\s*\[([^\s-]+)(-(.*))?\]\s*$/)
    {
      $section = $1;
      $sub_section = $3 || '';
    }
    elsif (/^\s*\`(.*)\`\s*$/)
    {
      my %cmd;
      $cmd{'section'} = $section;
      $cmd{'sub_section'} = $sub_section;
      $cmd{'backticked'} = $1 || '';
      push @lines, \%cmd;
    }
    else
    {
      my $line = $_;
      chomp($line);
      @line = &split_line($line);

      my $parser = 'parse_' . $section;

      if (defined(&$parser))
      {
	my $cmdref = &$parser(@line);

	if (defined($cmdref))
	{
	  $$cmdref{'section'} = $section;
	  $$cmdref{'sub_section'} = $sub_section;

	  # Set the revision to HEAD if the one specified in the prodspec
	  # file is to be ignored
	  if ($ignore_revision && defined($$cmdref{'revision'}))
	  {
	    if (defined($$cmdref{'branch'}))
	    {
	      $$cmdref{'revision'} = $$cmdref{'branch'};
	    }
	    else
	    {
	      $$cmdref{'revision'} = 'HEAD';
	    }
	  }

	  push @lines, $cmdref;
	}
      }
      else
      {
	print("Do not know how to handle $section section for: $line\n");
      }
    }
  }

  close(PRODSPEC);

  return @lines;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_shell
##
##****************************************************************************

sub parse_shell
{
  my(@line) = @_;
  my %cmd;

  $cmd{'text'} = &join_line(@line);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_config
##
##****************************************************************************

sub parse_config
{
  my(@line) = @_;
  my %cmd;

  $cmd{'option'} = shift @line;
  $cmd{'value'}  = shift @line;

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_release
##
##****************************************************************************

sub parse_release
{
  my(@line) = @_;
  my %cmd;

  $cmd{'text'} = &join_line(@line);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_build
##
##****************************************************************************

sub parse_build
{
  my(@line) = @_;
  my %cmd;

  $cmd{'text'} = &join_line(@line);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_jimageid
##
##****************************************************************************

sub parse_jimageid
{
  my(@line) = @_;
  my %cmd;

  $cmd{'text'} = &join_line(@line);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_tools
##
##****************************************************************************

sub parse_tools
{
  my(@line) = @_;
  my %cmd;

  if ($line[0] =~ /^(R\S+)/)
  {
    $cmd{'file'}      = 'tools';
    $cmd{'full_file'} = 'tools';
    $cmd{'revision'}  = $1;
    return \%cmd;
  }

  return undef;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_source
##
##****************************************************************************

sub parse_source
{
  my(@line) = @_;
  my %cmd;

  $cmd{'file'} = shift @line;
  $cmd{'full_file'} = $cmd{'file'};

  &parse_standard_options(\@line, \%cmd);

  $cmd{'options'} = &join_line(@line);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_packages
##
##****************************************************************************

sub parse_packages
{
  my(@line) = @_;
  my %cmd;

  $cmd{'file'} = shift @line;
  $cmd{'full_file'} = 'packages/' . $cmd{'file'};

  &parse_standard_options(\@line, \%cmd);

  $cmd{'options'} = &join(@line);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_devs
##
##****************************************************************************

sub parse_devs
{
  my(@line) = @_;

  # Special treatment of 'standard' and 'scsi'

  return undef;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_libs
##
##****************************************************************************

sub parse_libs
{
  my(@line) = @_;
  my %cmd;

  # Special treatment of 'glibc'

  $cmd{'file'} = shift @line;
  $cmd{'full_file'} = 'libs/' . $cmd{'file'};

  &parse_standard_options(\@line, \%cmd);

  $cmd{'options'} = &join_line(@line);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_apps
##
##****************************************************************************

sub parse_apps
{
  my(@line) = @_;
  my %cmd;

  $cmd{'file'} = shift @line;
  $cmd{'full_file'} = 'apps/' . $cmd{'file'};

  &parse_standard_options(\@line, \%cmd);

  $cmd{'options'} = &join_line(@line);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_dirs
##
##****************************************************************************

sub parse_dirs
{
  my(@line) = @_;
  my %cmd;

  $cmd{'directory'} = shift @line;

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_files
##
##****************************************************************************

sub parse_files
{
  my(@line) = @_;
  my %cmd;

  $cmd{'file'} = shift @line;

  &parse_file_options(\@line, \%cmd);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_links
##
##****************************************************************************

sub parse_links
{
  my(@line) = @_;
  my %cmd;

  $cmd{'link'} = shift @line;
  $cmd{'destination'} = shift @line;

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_swinfo
##
##****************************************************************************

sub parse_swinfo
{
  my(@line) = @_;
  my %cmd;

  my($generator, $file) = split(':', $line[0]);
  $cmd{'generator'} = $generator;
  $cmd{'file'} = $file;

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_image
##
##****************************************************************************

sub parse_image
{
  my(@line) = @_;
  my %cmd;

  $cmd{'text'} = &join_line(@line);

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_mkfdscript
##
##****************************************************************************

sub parse_mkfdscript
{
  my(@line) = @_;

  # OBSOLETE

  return undef;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_kernel
##
##****************************************************************************

sub parse_kernel
{
  my(@line) = @_;

  # OBSOLETE

  return undef;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_ptable
##
##****************************************************************************

sub parse_ptable
{
  my(@line) = @_;
  my %cmd;

  $cmd{'name'} = $line[0];
  $cmd{'size'} = $line[1];
  $cmd{'rw'}   = $line[2];
  $cmd{'type'} = $line[3];
  $cmd{'file'} = $line[4];

  return \%cmd;
}

#*****************************************************************************
##
## FUNCTION NAME: parse_standard_options
##
##****************************************************************************

sub parse_standard_options
{
  my($lineref, $cmdref) = @_;

  for (my $i = 0; $i <= $#$lineref;)
  {
    if ($$lineref[$i] =~ /^-(.+)$/)
    {
      my $opt = $1;

      if ($opt =~ /^r(.*)/)
      {
	if (defined($1) && length($1))
	{
	  $$cmdref{'revision'} = $1;
	  splice(@$lineref, $i, 1);
	}
	elsif ($i == $#$lineref)
	{
	  $$cmdref{'revision'} = 'HEAD';
	  splice(@$lineref, $i, 1);
	}
	else
	{
	  $$cmdref{'revision'} = $$lineref[$i+1];
	  splice(@$lineref, $i, 2);
	}
      }
      elsif ($opt =~ /^branch/)
      {
	if ($i == $#$lineref)
	{
	  $$cmdref{'branch'} = 'HEAD';
	  splice(@$lineref, $i, 1);
	}
	else
	{
	  $$cmdref{'branch'} = $$lineref[$i+1];
	  splice(@$lineref, $i, 2);
	}
      }
      elsif ($opt =~ /^s(.*)/)
      {
	if (defined($1) && length($1))
	{
	  $$cmdref{'source'} = $1;
	  splice(@$lineref, $i, 1);
	}
	elsif ($i != $#$lineref)
	{
	  $$cmdref{'source'} = $$lineref[$i+1];
	  splice(@$lineref, $i, 2);
	}
      }
      elsif ($opt =~ /^noreleasenote$/)
      {
	$$cmdref{'noreleasenote'} = 1;
	splice(@$lineref, $i, 1);
      }
      elsif ($opt =~ /^target$/)
      {
	if ($i == $#$lineref)
	{
	  $$cmdref{'target'} .= '';
	  splice(@$lineref, $i, 1);
	}
	else
	{
	  $$cmdref{'target'} .= ' ' . $$lineref[$i+1];
	  splice(@$lineref, $i, 2);
	}
      }
      elsif ($opt =~ /^notarget$/)
      {
	$$cmdref{'target'} = '';
	splice(@$lineref, $i, 1);
      }
      elsif ($opt =~ /^noautotarget$/)
      {
	$$cmdref{'noautotarget'} = 1;
	splice(@$lineref, $i, 1);
      }
      else
      {
	$i++;
      }
    }
    else
    {
      $i++;
    }
  }

  $$cmdref{'target'} = 'install' unless (defined($$cmdref{'target'}));
}

#*****************************************************************************
##
## FUNCTION NAME: parse_file_options
##
##****************************************************************************

sub parse_file_options
{
  my($lineref, $cmdref) = @_;

  for (my $i = 0; $i <= $#$lineref;)
  {
    if ($$lineref[$i] =~ /^-(.+)$/)
    {
      my $opt = $1;

      if ($opt =~ /^d(.*)/)
      {
	if (defined($1) && length($1))
	{
	  $$cmdref{'dest'} = $1;
	  splice(@$lineref, $i, 1);
	}
	elsif ($i != $#$lineref)
	{
	  $$cmdref{'dest'} = $$lineref[$i+1];
	  splice(@$lineref, $i, 2);
	}
      }
      elsif ($opt =~ /^u(.*)/)
      {
	if (defined($1) && length($1))
	{
	  $$cmdref{'uid'} = $1;
	  splice(@$lineref, $i, 1);
	}
	elsif ($i != $#$lineref)
	{
	  $$cmdref{'uid'} = $$lineref[$i+1];
	  splice(@$lineref, $i, 2);
	}
      }
      elsif ($opt =~ /^g(.*)/)
      {
	if (defined($1) && length($1))
	{
	  $$cmdref{'gid'} = $1;
	  splice(@$lineref, $i, 1);
	}
	elsif ($i != $#$lineref)
	{
	  $$cmdref{'gid'} = $$lineref[$i+1];
	  splice(@$lineref, $i, 2);
	}
      }
      elsif ($opt =~ /^p(.*)/)
      {
	if (defined($1) && length($1))
	{
	  $$cmdref{'perm'} = $1;
	  splice(@$lineref, $i, 1);
	}
	elsif ($i != $#$lineref)
	{
	  $$cmdref{'perm'} = $$lineref[$i+1];
	  splice(@$lineref, $i, 2);
	}
      }
    }
    else
    {
      $i++;
    }
  }
}

#*****************************************************************************
##
## FUNCTION NAME: split_line
##
##****************************************************************************

sub split_line
{
  my($line) = @_;
  my(@lines) = parse_line('\s+', 0, $line);

  # For some strange reason parse_lines() adds an undef'ed entry at the end
  pop(@lines) if (!defined($lines[$#lines]));

  return @lines;
}

#*****************************************************************************
##
## FUNCTION NAME: join_line
##
##****************************************************************************

sub join_line
{
  my(@line) = @_;

  return join(' ', map { if (/\s/) { '"' . $_ . '"' } else { $_ } } @line);
}

#*****************************************************************************

# Always end a package with a true value (to mark successful loading).
1;

#****************** END OF FILE prodspec.pm **********************************
